# -*- encoding: utf-8 -*-
__version__ = '0.0.17'
DEFAULT_PORT = 5672
AMQP_EXCHANGE = "amqppy"
AMQP_BROKER = "amqp://localhost:{}//".format(DEFAULT_PORT)

import logging
try:
    # not available in python 2.6
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):

        def emit(self, record):
            pass

# Add NullHandler to prevent logging warnings
logging.getLogger(__name__).addHandler(NullHandler())


class RpcRemoteException(Exception):
    '''
    It would be raised in the publisher.Rpc.request() when remote reply fails
    '''
    pass


class ResponseTimeout(Exception):
    '''
    It would be raised in the publisher.Rpc.request() when remote reply exceeds its allowed execution time, the timeout.
    '''
    pass


class PublishNotRouted(Exception):
    '''
    It would be raised in the publisher.Rpc.request() or publisher.Topic.publish() when there is no consumer listening
    those Topics or Rpc requests.
    '''
    pass


class ExclusiveQueue(Exception):
    '''
    It would be raised in the consumer.Worker.add_topic() or consumer.Worker.add_request() when tries to consume from a queue
    where there is already a consumer listening. That happens when add_topic or add_request is called with 'exclusive=True'.
    '''
    pass


class ExchangeNotFound(Exception):
    '''
    It will be raised when AMQP Exchange does not exist.
    '''
    pass


class AbortConsume(Exception):
    '''
    This exception can be raised by the Topic callback or RPC reply callback. And indicates to amqppy to do not send ACK
    for that consuming message. For this, is required 'no_ack=False' in consumer.Worker.add_topic() or consumer.Worker.add_request()
    '''
    pass


class DeadLetterMessage(Exception):
    '''
    This exception can be raised by the Topic callback or RPC reply callback. And indicates to amqppy to move this
    message is being consumed to the DeadLetter Queue. See: 'https://www.rabbitmq.com/dlx.html'
    '''
    pass



# -*- encoding: utf-8 -*-
import sys
import os
import pika
import uuid
import time
import json
import logging
# add amqppy path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import amqppy
from amqppy import utils


logger = logging.getLogger(__name__)


class Topic(object):
    """This class creates a connection to the message broker and provides a method to publish messages on
    one topic also known as routing_key in AMQP terms. The class instance will create on single connection
    to the broker for all the topic published.

    :param str broker: The URL for connection to RabbitMQ. Eg: 'amqp://serviceuser:password@rabbit.host:5672//'
    """
    def __init__(self, broker):
        self._connection = utils._create_connection(broker=broker)

    def __del__(self):
        logger.debug("publisher destructor")
        if self._connection and self._connection.is_open:
            logger.debug("closeing connection")
            self._connection.close()

    def publish(self, exchange, routing_key, body, headers=None, persistent=True):
        """Publish a message to the given exchange and a routing key.

        :param str exchange: The exchange you want to publish the message.
        :param str rounting_key: The rounting key to bind on
        :param str body: The body of the message you want to publish. It should be of type unicode or string encoded with UTF-8.
        :param dict headers: Message headers.
        :param bool persistent: Makes message persistent. The message would not be lost after RabbitMQ restart.
        """
        logger.debug("creating channel")
        channel = self._connection.channel()
        try:
            channel.confirm_delivery()
            logger.debug("publishing message at exchange: {} and routing_key: {}".format(exchange, routing_key))
            publish_result = channel.basic_publish(exchange=exchange,
                                                   routing_key=routing_key,
                                                   properties=pika.BasicProperties(
                                                       content_encoding='utf-8',
                                                       delivery_mode=2 if persistent else 1, headers=headers,  # 2 -> persistent
                                                   ),
                                                   body=utils._ensure_utf8(body),
                                                   mandatory=True)  # to know if the message was routed
            if not publish_result:
                logger.debug("Publisher published message was not routed")
                raise amqppy.PublishNotRouted("Publisher published message was not routed")
        except pika.exceptions.ChannelClosed as e:
            if "NOT_FOUND - no exchange" in str(e):
                raise amqppy.ExchangeNotFound(str(e))
            else:
                raise e

        finally:
            if channel and channel.is_open:
                logger.debug("closing channel")
                channel.close()


class Rpc(object):
    """The duty of Rpc class is to make RPC requests and returns its responses.
    `RPC pattern tutorial <https://www.rabbitmq.com/tutorials/tutorial-six-python.html>`_.
    The class instance will create on single connection to the broker for all the RPC requests.

    :param str broker: The URL for connection to RabbitMQ. Eg: 'amqp://serviceuser:password@rabbit.host:5672//'
    """
    def __init__(self, broker):
        self._connection = utils._create_connection(broker=broker)

    def __del__(self):
        logger.debug("rpc destructor")
        if self._connection and self._connection.is_open:
            # logger.debug("closeing connection")
            self._connection.close()

    def _on_response(self, ch, method, props, body):
        if self.corr_id == props.correlation_id:
            if isinstance(body, bytes):
                body = body.decode("utf-8")
            if not isinstance(body, str):
                logger.warning("_on_response, type: {}, body: {}".format(type(body), body))
            logger.debug("_on_response body: {}".format(body))
            self.response = json.loads(body)

    def request(self, exchange, routing_key, body, timeout=10.0):
        """Makes a RPC request and returns its response.
        `RPC pattern tutorial <https://www.rabbitmq.com/tutorials/tutorial-six-python.html>`_.
        This call creates and destroys a connection every time, if you want to save connections, please use the class Rpc.

        :param str rounting_key: The routing key to bind on
        :param str body: The body of the message request you want to request. It should be of type unicode or string encoded with UTF-8.
        :param str exchange: The exchange you want to publish the message.
        :param bool timeout: Maximum seconds to wait for the response.
        """
        self.exchange = exchange
        channel = self._connection.channel()
        # Enabled delivery confirmations:
        # very important to know if the message was delivered or consumed
        channel.confirm_delivery()

        try:
            # rpc response queue
            self.response_queue = channel.queue_declare(exclusive=True).method.queue
            channel.queue_bind(queue=self.response_queue, exchange=self.exchange, routing_key=self.response_queue)

            # lets listen response
            channel.basic_consume(queue=self.response_queue, consumer_callback=self._on_response, no_ack=True)

            logger.debug("publishing rpc request, exchange: {}, routing_key: {}, body: {}".format(self.exchange, routing_key, body))
            self.response = None
            self.corr_id = str(uuid.uuid4())
            publish_result = channel.basic_publish(exchange=self.exchange,
                                                   routing_key=routing_key,
                                                   properties=pika.BasicProperties(
                                                       reply_to=self.response_queue,
                                                       correlation_id=self.corr_id,
                                                       content_encoding='utf-8',
                                                       delivery_mode=1),  # 2 -> persistent
                                                   body=utils._ensure_utf8(body),
                                                   mandatory=True)
            if not publish_result:
                logger.debug("Rpc published message was not routed")
                raise amqppy.PublishNotRouted("Rpc published message was not routed")

            # wait for response
            timeout = max(timeout, 0.1)
            logger.debug("waiting for rpc response... on \'{}\' for {} seconds".format(self.response_queue, timeout))
            self._connection.process_data_events(timeout)
            if not self.response:
                logger.warning("Rpc Timeout has been triggered waiting for the response")
                raise amqppy.ResponseTimeout("Rpc Timeout has been triggered waiting for the response")
            """
            start = time.time()
            while self.response is None:
                self._connection.process_data_events(min(timeout, 1.0))
                if timeout > 0 and time.time() - start >= timeout:
                    logger.warning("AMQP RPC Timeout has been triggered waiting for the response")
                    raise amqppy.ResponseTimeout("AMQP RPC Timeout has been triggered waiting for the response")
            """
            # we have the rpc reply
            # is success?
            if "result" in self.response:
                return self.response["result"]
            else:
                str_error = "Unknown RPC error"
                if "error" in self.response:
                    str_error = self.response["error"]
                raise amqppy.RpcRemoteException(str_error)
        except pika.exceptions.ChannelClosed as e:
            if "NOT_FOUND - no exchange" in str(e):
                raise amqppy.ExchangeNotFound(str(e))
            else:
                raise e
        finally:
            if channel and channel.is_open:
                logger.debug("closing channel")
                channel.close()



# -*- encoding: utf-8 -*-
import traceback
from functools import wraps
import collections
# from collections import namedtupla
import pika
import logging
import threading
import time
import json
import sys
import os

# add amqppy path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import amqppy
from amqppy import utils


logger = logging.getLogger(__name__)

_ChannelExchange = collections.namedtuple('ChannelExchange', ['channel', 'exchange'])


class Worker(object):
    """ This class handles a worker that listens for incoming Topics and Rpc requests.

    :param str broker: The URL for connection to RabbitMQ. Eg: 'amqp://serviceuser:password@rabbit.host:5672//'
    """
    def __init__(self, broker, heartbeat_sec=None):
        # map(callback) -> (channel, exchange)
        self._callbacks = {}
        self.quit = False
        self.thread = None
        self._conn = None
        try:
            logger.debug("connecting to broker \'{}\'".format(broker))
            self._conn = utils._create_connection(broker=broker, heartbeat_sec=heartbeat_sec)
        finally:
            if self._conn and self._conn.is_open:
                logger.debug("connected")
            else:
                logger.error("cannot connect to broker \'{}\'".format(broker))

    def __del__(self):
        # logger.debug("consumer worker destructor")
        self._close()

    def _close(self):
        for callback in self._callbacks:
            if self._callbacks[callback].channel and self._callbacks[callback].channel.is_open:
                self._callbacks[callback].channel.close()
        self._callbacks = {}
        if self._conn and self._conn.is_open:
            logger.debug('closing connection')
            self._conn.close()
            self._conn = None
            logger.debug('connection closed')

    def _create_channel(self, exchange, callback):
        try:
            channel = self._conn.channel()
            channel.exchange_declare(exchange=exchange, exchange_type="topic", passive=True)
            self._callbacks[callback] = _ChannelExchange(channel, exchange)
            return channel
        except Exception:
            channel = self._conn.channel()
            channel.exchange_declare(exchange=exchange, exchange_type="topic", passive=False, durable=True, auto_delete=True)
            self._callbacks[callback] = _ChannelExchange(channel, exchange)
            return channel

    def stop(self):
        """ Stops listening and close all channels and the connection
        """
        logger.debug("stop")
        self.quit = True
        self._join()
        self._close()

    def add_request(self, routing_key, on_request_callback, exchange=amqppy.AMQP_EXCHANGE, durable=False, auto_delete=True,
                    exclusive=False):
        """ Registers a new consumer for a RPC reply task. These tasks will be executed when a RPC request is invoked by
        publisher.Rpc.request().

        :param str rounting_key: It defines the subscription interest. In terms of AMQP the routing key to bind on
        :param method on_request_callback: Called when a Rpc request is invoked, it should return the reply.
        :param str exchange: The exchange you want to publish the message.
        :param bool durable: Queue messages survives a reboot of RabbitMQ.
        :param bool auto_delete: Queues will auto-delete after use.
        :param bool exclusive: Ensures that is the unique consumer
        """
        logger.debug("adding request, exchange: {}, topic: {} --> {}".format(exchange, routing_key, on_request_callback))
        channel = self._create_channel(exchange, on_request_callback)
        channel.queue_declare(queue=routing_key, durable=durable, auto_delete=auto_delete)
        channel.queue_bind(queue=routing_key, exchange=exchange, routing_key=routing_key)
        channel.confirm_delivery()
        try:
            channel.basic_consume(
                exclusive=exclusive,
                queue=routing_key,
                consumer_callback=self._profiler_wrapper_request(on_request_callback),
                no_ack=True)
        except pika.exceptions.ChannelClosed as e:
            if "in exclusive use" in str(e):
                raise amqppy.ExclusiveQueue(str(e))
            else:
                raise e

        return self  # Fluent pattern

    def _profiler_wrapper_request(self, on_request_callback):
        @wraps(on_request_callback)
        def _wrapper(*args, **kwargs):
            logger.debug("request \'{}\'.*args: {}".format(on_request_callback.__name__, args))
            # process request arguments
            deliver = args[1]
            properties = args[2]
            message = args[3]
            # convert message body to string
            if isinstance(message, bytes):
                message = message.decode("utf-8")
            if not isinstance(message, str):
                logger.warning("_profiler_wrapper_request, type: {}, body: {}".format(type(message), message))

            logger.debug("Starting request \'{}\'".format(on_request_callback.__name__))
            # response = on_request_callback(*args, **kwargs)
            start = time.time()
            try:
                response = {
                    # message is text, it should be converted in dictionary at request func
                    "result": on_request_callback(exchange=deliver.exchange, routing_key=deliver.routing_key, headers=properties.headers, body=message),
                }
            except Exception as e:
                logger.warning("Exception in request \'{}\', routing_key: {}\n{}".format(on_request_callback.__name__,
                                                                                         deliver.routing_key,
                                                                                         traceback.format_exc()))
                response = {
                    "error": str(e, encoding='utf8')
                }
            elapsed = time.time() - start
            logger.debug('Request \'{}\' finished. Time elapsed: {}'.format(on_request_callback.__name__, elapsed))

            # sending response back
            channel = self._callbacks[on_request_callback].channel
            exchange = self._callbacks[on_request_callback].exchange
            routing_key = properties.reply_to
            logger.debug('Sending RPC response to routing key: {}'.format(routing_key))
            try:
                publish_result = channel.basic_publish(
                    exchange=exchange,
                    routing_key=routing_key,
                    properties=pika.BasicProperties(
                        correlation_id=properties.correlation_id,
                        content_encoding='utf-8',
                        content_type='application/json'),
                    body=json.dumps(response, ensure_ascii=False).encode('utf8'),
                    mandatory=True)
                if not publish_result:
                    raise amqppy.PublishNotRouted("Request response was not routed")
            except amqppy.PublishNotRouted:
                # don't raise it
                logger.warning("RPC response it has not been published, it might be due to a response waiting timeout")
            except Exception:
                logger.error('Exception on publish message to routing_key: {}. Exception message: {}'.format(
                    routing_key, traceback.format_exc()))
            logger.debug('RPC response sent.')
        return _wrapper

    def add_topic(self, routing_key, on_topic_callback, queue=None, exclusive=False, exchange=amqppy.AMQP_EXCHANGE, durable=False,
                  auto_delete=True, no_ack=True, **kwargs):
        """ Registers a new consumer for a Topic subscriber. These tasks will be executed when a Topic is published by
        publisher.Topic.publish().

        :param str rounting_key: The routing key to bind on.
        :param method on_topic_callback: Called when a topic is published.
        :param str queue: The name of the queue. If it is not provided the queue will be named the same as the 'routing_key'.
        :param bool exclusive: Only one consumer is allowed.
        :param str exchange: The exchange you want to publish the message.
        :param bool durable: Queue messages survives a reboot of RabbitMQ.
        :param bool auto_delete: Queues will auto-delete after use.
        :param bool no_ack: Tell the broker that ACK reply is not needed. If it is False, an ACK will be sent automatically each \
        time a message is consumed unless a amqppy.AbortConsume or amqppy.DeadLetterMessage is raised.
        """
        logger.debug("adding topic, exchange: {}, topic: {} --> {}".format(exchange, routing_key, on_topic_callback, kwargs))
        self.no_ack = no_ack
        channel = self._create_channel(exchange, on_topic_callback)
        queue_name = queue if queue else routing_key
        channel.queue_declare(queue=queue_name, durable=durable, auto_delete=auto_delete,
                              arguments=kwargs)
        channel.queue_bind(queue=queue_name, exchange=exchange, routing_key=routing_key)
        try:
            channel.basic_consume(
                queue=queue_name,
                exclusive=exclusive,
                consumer_callback=self._profiler_wrapper_topic(on_topic_callback),
                no_ack=no_ack)
        except pika.exceptions.ChannelClosed as e:
            if "in exclusive use" in str(e):
                raise amqppy.ExclusiveQueue(str(e))
            else:
                raise e
        return self  # Fluent pattern

    def _profiler_wrapper_topic(self, on_topic_callback):
        @wraps(on_topic_callback)
        def _wrapper(*args, **kwargs):
            logger.debug("topic \'{}\'.*args: {}".format(on_topic_callback.__name__, args))
            # logger.debug("request \'{}\'.**kwargs: {}".format(on_topic_callback.__name__, kwargs))
            # process request arguments
            deliver = args[1]
            properties = args[2]
            message = args[3]
            # convert message body to string
            if isinstance(message, bytes):
                message = message.decode("utf-8")
            if not utils._is_string(message):
                logger.warning("_profiler_wrapper_topic, message in not string, is of type: {}".format(type(message)))
            # logger.debug("Properties vars: {}".format(vars(properties)))
            logger.debug("Starting request \'{}\'".format(on_topic_callback.__name__))
            start = time.time()
            try:
                on_topic_callback(exchange=deliver.exchange, routing_key=deliver.routing_key, headers=properties.headers, body=message)
                if not self.no_ack:
                    self._callbacks[on_topic_callback].channel.basic_ack(delivery_tag=deliver.delivery_tag)
                    logger.debug("ACK sent")
            except amqppy.AbortConsume as e:
                logger.warning("AbortConsume exception: {}".format(e))
            except amqppy.DeadLetterMessage as e:
                logger.warning("DeadLetterMessage exception: {}".format(e))
                self._callbacks[on_topic_callback].channel.basic_reject(delivery_tag=deliver.delivery_tag, requeue=False)
            finally:
                elapsed = time.time() - start
                logger.debug('Request \'{}\' finished. Time elapsed: {}'.format(on_topic_callback.__name__, elapsed))

        return _wrapper

    def run(self):
        """ Start worker to listen. This will block the execution until the worker is stopped or an uncaught Exception  """
        logger.info('Running worker, waiting for the first message...')
        try:
            while not self.quit:
                self._conn.process_data_events(0.5)
        finally:
            logger.info("exiting from worker run")

    def run_async(self):
        """ Start asynchronously worker to listen. The execution thread will follow after this call, hence is not blocked.  """
        self.thread = threading.Thread(target=self.run)
        self.thread.start()
        return self  # Fluent pattern

    def _join(self):
        """ Waits until worker has ended """
        if self.thread and self.thread.is_alive():
            self.thread.join()
